
REVISION -aana "0.5G -- aana.frt : Immediate request for buffer ---"

PRIVATES
2 VALUE PORT#
\ Analyse a transputer that is hanging of DBLINK-IN DBLINK-OUT .
\ Works together with a asswrite.frt that must preceed this.
\ It uses variables present in asswrite.frt.
: DBLINK-IN LINK0-IN  PORT# CELLS + ;   PRIVATE
: DBLINK-OUT LINK0-OUT PORT# CELLS + ;   PRIVATE

\ Returns remainder regardless of N/D fits in a single
: MMOD  ( double N, single D -- remainder N modulo D )
    LOCAL D
    D MOD D FM/MOD DROP
;

: RESET
  S" PORT# TO PORT /s 2 /r TFORTH" EVALUATE
; PRIVATE

: StopForAnalyse
    RESET
;
\ `a' is an signed addres in the transputer that is being analysed.
\ Return `c' the content of that address.
: TC@           ( INT a -- UINT c)
    1 DBLINK-OUT CHANNEL-C!
    ( a) DBLINK-OUT CHANNEL-!
    DBLINK-IN CHANNEL-@
;

\ `a' is an offset in the workspace in the transputer that is being analysed.
\ Return `c' the address.
: WS            ( INT a -- UINT c)
    CELLS WorkSpaceStart +
;

\ `a' is an offset in the workspace in the transputer that is being analysed.
\ Return `c' the content of that address.
: WS@           ( INT a -- UINT c)
    CELLS WorkSpaceStart + TC@
;
\ `a' is an offset in the workspace in the transputer that is being analysed.
\ Return `c' the content of that address.
: DWS@           ( INT a -- DUINT c)
    CELLS WorkSpaceStart + DUP TC@ SWAP CELL+ TC@
;

\ Display the content of the FWORD or WWORD n
: WS?            ( FWORD/WWORD n -- )
    WS@
    DUP . H.
;

\ Display the content of the DWORD `d'
: DWS?           ( DWORD d -- )
    DWS@
    2DUP D. H. H.
;

\ Analyse code from address `a', assuming it is just loaded.
: TIDIS                 ( a -- )
   CodeStart -
   CodeSpace + IDIS
;

VARIABLE WHERE

: //
  WHERE @ TC@ H.
  1 CELLS WHERE +!
;

: ??
  WHERE @ TC@ H.
  1 CELLS WHERE +!
;

CREATE OFFSETS
7   , \   0
9   , \   1
19  , \   2
21  , \   3
37  , \   4
39  , \   5
49  , \   6
51  , \   7
61  , \   8
63  , \   9
67  , \  10
69  , \  11
79  , \  12
81  , \  13
91  , \  14
93  , \  15
97  , \  16
99  , \  17
109 , \  18
111 , \  19
121 , \  20
123 , \  21
139 , \  22
141 , \  23
151 , \  24
153 , \  25
181 , \  26
183 , \  27
187 , \  28
189 , \  29

DOC
                DUP PRIME? 0= IF
                   . ." is not prime"
                ELSE
                   DROP
                THEN
ENDDOC

\ Returns some divider of `n' or 0 if N is prime.
: FAC   ( double N -- int d )
    DLOCAL N
    2 LOCAL D
    BEGIN
      N D D M* D< 0= WHILE
      N D MMOD 0= IF D EXIT THEN
      1 +TO D
\       .S
    REPEAT
    0
;

: SHOW-PROBLEM
    CR 2DUP D. FAC .
;

: find-errors
    HighPointOfSieve WS@ LOCAL END        \ But lower address than START
    the-sieve WS@ LOCAL START
    NumberStart DWS@ DLOCAL Number

    BEGIN
        START TC@
        PPBY 0 DO
              \ Use 0= to "harden" the flags before comparing them.
              DUP 1 AND 0=
              Number   OFFSETS I CELLS + @   S>D D+    FAC 0= 0=
              <> IF
                  Number   OFFSETS I CELLS + @   S>D D+
                  SHOW-PROBLEM
              THEN
            1 RSHIFT
        LOOP DROP
        -1 CELLS +TO START
        NPE S>D +TO Number
        &. EMIT
    END START > KEY? OR UNTIL
;

\ Returns zero if the code down loaded has still the correct crc.
: PostMortemCrc
    PAD
    I_CodeEnd WS@ I_CodeStart WS@ DO
       I TC@
       OVER ! 1 CELLS +
   4 +LOOP

  \ The end of the area is now on the stack
  1 CELLS - DUP >R      \ Save address of crc
  PAD SWAP CrcOfArea
  R> @ -        \ Compare with stored crc.
;

: compare
    PostMortemCrc IF
        I_CodeStart WS@ CodeSpace
        BEGIN
               OVER TC@ OVER @ <> IF
                CR OVER H. SPACE OVER TC@ H. SPACE DUP @ H.
                KEY &Q = IF QUIT THEN
               THEN
            1 CELLS + SWAP   1 CELLS + SWAP
        OVER I_CodeEnd WS@ = UNTIL
    THEN
;

StopForAnalyse
\ Prevent errors by trying a calculating
: CALC
   CR ." You can't CALC, analysis is going on"
;
: CALC-RANGE
   CR ." You can't CALC_RANGE, analysis is going on"
;
DEPRIVE
