\ COPYRIGHT 1992@1997 DUTCH FORTH WORKSHOP - FILENAME : DISPLAY.FRT #4

NEEDS -graphics
NEEDS -auxtwin

REVISION -display  " DISPLAY  1.0F --- display : scrolling --"


DECIMAL
\ Gives an auditory signal
: BEEP 7 EMIT ;


\ Assumes a lot of variables are present.
\ Output screen looks as follows

\  0 title
\  1 current job running
\  2 status twins & range covered
\  3 what prime batches
\  4 and following : start for transputers
\ Constants for screen layout:
3 CONSTANT #COLUMNS
C/L #COLUMNS / CONSTANT #CW         \ Column Width
2  CONSTANT #BAR        \ Line number with the bar
4  CONSTANT #PR         \ Line with Partial Result
7  CONSTANT #SEP        \ Line with separation (one pixel horizontal line)
8  CONSTANT #LT         \ Line with Table of range assigned to each transputer
17 CONSTANT #LTD        \ Number of lines available for transputer display.
29 CONSTANT #LL         \ Last line, used for debugging and errors
200 CONSTANT #MC        \ Maximum length of chain

    \ twinhelp may need the layout above.
    NEEDS -twinhelp

FALSE VALUE CONNECTED

\                  AUXILIARY GRAPHICS

\ Draw a vertical separation at the begin of column L
: VERT-SEPARATION    ( int L -- )
    #CW * #CHW * LOCAL X
    X    Ymax #LT #CH * - #CH 2/ +
    X    Ymax #LL #CH * - #CH 2/ +
    white DRAW-LINE
;

\ Draw a separation line at (character line) L
: HOR-SEPARATION    ( int L -- )
    Ymax SWAP #CH * - #CH 2/ - LOCAL Y
    0 Y #CW #CHW * #COLUMNS * Y white DRAW-LINE
;

: START-MESSAGE
    CONNECTED 0= IF EXIT THEN
    {{ 23 0 AT-XY ." >>> PRIME NUMBER TWINS COUNTING <<<" }}
    {{
        0 1 AT-XY ." Sieving between "
        N-START 2@ SPACE DP. ." and" N-END 2@ SPACE DP.
    }}
;

: THERMOMETER-SCALE

    RECORD-ATTEMPT @ 0= IF
        0 #BAR AT-XY ." No record attempt"
        EXIT
    THEN

    \ Define scaled versions of doubles needed.
    CURRENT-RECORD 2@ 1G-SCALE LOCAL CURRENT-RECORD'
    OUR-RECORD     2@ 1G-SCALE LOCAL OUR-RECORD'
    CHUNKS            1G-SCALE LOCAL CHUNKS'
    CURRENT-START  2@ 1G-SCALE LOCAL CURRENT-START'
    CURRENT-START 2@ N-START 2@ D- 1G-SCALE LOCAL CURRENT-OFFSET'

    Ymax #BAR 2+ #CH * - 1+ LOCAL Y        \ Offset to lower end of bar

    \ Marks the ends of the scale and the place of the current record
    0 Y     2DUP #CH 2* + green DRAW-LINE
    Xmax Y  2DUP #CH 2* + green DRAW-LINE
    \ Draw the neutral scale
    0 Y #CH 2/ +    Xmax #CH  green FILLED-RECTANGLE
    \ A red line where the record is
    Xmax CURRENT-RECORD' OUR-RECORD' */ Y
    2DUP #CH 2* + red DRAW-LINE
    \ Draw the active scale
    0
    Y #CH 2/ + 1+
    Xmax CURRENT-START' OUR-RECORD' */
    #CH 2-
    red FILLED-RECTANGLE

    \ Draw a single line with the current progress
    0   Y   Xmax   Y   green   DRAW-LINE
    0   Y   Xmax CURRENT-OFFSET' CHUNKS' */   Y   red   DRAW-LINE
;

\ Place the cursor at the print start of column COL
: AT-COL        ( int COL -- )
   #CW * 3 +
;

\ Show how far we are.
: .STATUS       ( double CURRENT-START -- )
    CONNECTED 0= IF 2DROP EXIT THEN
    {{ THERMOMETER-SCALE }}
    {{  0 #PR 1+ AT-XY
       ?DONE #PENDING 0= AND IF ." Final " ELSE ." Partial " THEN
       ." result after "   .ELAPSED ." :"
       0 #PR 2+ AT-XY
       ." Up till" ( CURRENT-START) SPACE DP.
       ." we have accumulated"   TOTAL-TWINS 2@ SPACE DP.   ." twin-primes"
    }}
;

\ #COLUMNS by #MC double cells, see also STARTS[,]
CREATE STARTS #COLUMNS #MC * 2* CELLS ( HERE OVER ERASE) ALLOT

\ Address where the start of the table in machine M hanging of chain C+1
\ is stored.
\ A range starting at 0 is impossible, so it is used to indicate that
\ a machine is not currently used.
: STARTS[,]     ( int M, int C -- address START)
  #MC * + STARTS SWAP 2* CELLS +
;

\ The prime starts as present in the top of each column (counting from zero).
CREATE PRIME-STARTS HERE #COLUMNS CELLS DUP ALLOT ERASE

: .START-HEAD           ( int COL -- )
    LOCAL COL

    \ Chain is one more than column
    {{
       COL AT-COL #LT    AT-XY ." Chain " COL 1+ . ." : "
       COL 1+ CONFIGURATION[] DUP >N-MACHINE . ." with "
       >MEM 4 1000000 */ . ." Mb"
    }}
    PRIME-STARTS COL CELLS + @
    {{ COL AT-COL #LT 1+ AT-XY ." Primes at " 8 .R }}
    {{ COL AT-COL #LT 2 + AT-XY ." Transputer    Start" }}
;

\      NEEDS -assert
\ FIRST-SHOWN is the number of the machine presented uppermost in the
\ screen window for the Start's. This is the highest number (nearest to the
\ main computer.
CREATE FIRST-SHOWN #COLUMNS CELLS ALLOT
CREATE COL-CURSOR  #COLUMNS CELLS ALLOT
2VARIABLE THE-CURSOR    \ filled by ROW COL 2!
: INIT-CURSOR
    #COLUMNS 0 DO
        I 1+ CONFIGURATION[] >N-MACHINE
        1 - 0 MAX
        DUP FIRST-SHOWN I CELLS + !
        COL-CURSOR I CELLS + !
    LOOP
    0 0 THE-CURSOR 2!
;

\ If the cursor is off the screen, adjust FIRST-SHOWN such that it is
\ visible. If impossible, give a beep.
: ADJUST-FIRST-SHOWN
{
    THE-CURSOR 2@ LOCAL COL
    DUP -1 = IF
        0 COL THE-CURSOR 2! \ Can't scroll up
        FIRST-SHOWN COL CELLS + @ 1+
        COL 1+ CONFIGURATION[] >N-MACHINE
        = IF
            BEEP
        ELSE
            1 FIRST-SHOWN COL CELLS + +!
        THEN
    ELSE DUP #LTD = IF
        #LTD 1-  COL THE-CURSOR 2! \ Can't scroll down
        FIRST-SHOWN COL CELLS + @
        0= IF
            BEEP
        ELSE
            -1 FIRST-SHOWN COL CELLS + +!
        THEN
    THEN THEN DROP
}
;
\ Show the cursor if appropriate
: ?SHOW-CURSOR  ( int R,C -- )
    2DUP
        AT-COL #CW + 5 -  SWAP
        #LT + 3 +
        AT-XY
    THE-CURSOR 2@ D= IF &< ELSE BL THEN
    EMIT
;

\ Prints the one start value for the transputers that has to be
\ printed in column `C' and position `R' relative in the display.
\ This depend on scrolling as defined via FIRST-SHOWN.
\ ( FIRST-SHOWN is associated with position 0 relative in the display.)
: .ONE-START    ( int C,R -- )
    LOCAL ROW
    LOCAL COL
    \ Find out what number the machine for this position has.
    FIRST-SHOWN COL CELLS + @ ROW - LOCAL MACHINE

    {{
        COL AT-COL ROW #LT 3 +  + AT-XY
        MACHINE 0 < IF
            4 SPACES
        ELSE
            MACHINE 3 .R
        THEN
        MACHINE COL STARTS[,] 2@
        2DUP 0. D=    MACHINE 0 <   OR IF
            2DROP
            #CW 8 -  SPACES
        ELSE
            SPACE DP.
        THEN
        ROW COL ?SHOW-CURSOR
    }}
;

\ Remember the start `L' for machine `MACHINE' hanging off chain `CH'.
\ `CH' is an index into `CONFIGURATION' .
: REMEMBER-START    ( double L, int MACHINE, int CH -- )
    1- LOCAL COL
    LOCAL MACHINE

    MACHINE COL STARTS[,] 2!
    CONNECTED 0= IF EXIT THEN
    COL FIRST-SHOWN @ MACHINE - .ONE-START
;

\ Erase all the starts for chain `CH'
: ERASE-STARTS          ( int CH -- )
    1- LOCAL COL
    #MC 0 DO
        0. I COL STARTS[,] 2!
    LOOP
    CONNECTED IF
        #LTD 0 DO
            COL I .ONE-START
        LOOP
    THEN
;

1 ERASE-STARTS   2 ERASE-STARTS   3 ERASE-STARTS

\ Show the start of the prime table `N' for chain `CH'.
\ `CH' is an index into `CONFIGURATION' .
: REMEMBER-PRIMES ( int N, int CH -- )
    1-   \ Column 0 would otherwise be wasted
    2DUP CELLS PRIME-STARTS + !         \ Remember
    CONNECTED 0= IF DROP DROP EXIT THEN
    {{ AT-COL #LT 1+ AT-XY ." Primes at " 8 .R }}
;

\ Print the machine identification.
\ Offset by multiples of say 100 (#MC) for each chain.
: .MACHINE-ID
    MACHINE-ID @ N-CHANNEL #MC * + .
;

\ Show latest result.
: .LAST-ACCUMULATED
    CONNECTED 0= IF EXIT THEN
    {{
        0 #PR AT-XY
         ." Twin-primes accumulated from latest message: "
        N-TWIN-OUT @ .
    }}
;


\ Show all the starts as remembered
: .STARTS
    0 LOCAL COL

    {{ #LT 1- HOR-SEPARATION }}
    {{ #LL 1- HOR-SEPARATION }}
    #COLUMNS 1+ 0 DO I VERT-SEPARATION LOOP
\     PAGE .S FREEZE
    #LINKS 1- 0 ?DO     \ Not counting command link
        I .START-HEAD
        #LTD 0 DO
            J I .ONE-START
        LOOP
    LOOP
;
\ The transputer sent a debug message, only during debugging.
: .DEBUG
    CONNECTED 0= IF EXIT THEN
    {{
       0 #LL AT-XY ." DEBUG " DEBUG-RESULT @ DUP . H.
       ." ON " .MACHINE-ID
    }}
;


\ This is serious: an antibugging error.
: .LOG-ERROR
    DEBUG-RESULT @   MACHINE-ID @   TWIN-ERROR
    CONNECTED 0= IF EXIT THEN
    {{
        0 #LL AT-XY ." ERROR #" DEBUG-RESULT ?
        ." ON " .MACHINE-ID SPACE
    }}
;

\ Display the content of the array ERRORS.
: .SHOW-ERRORS
   {{
   ERRORS
   DUP CURRENT-ERROR = IF
      CR ." No errors"
   ELSE
       CR ." MACHINE    ERROR NUMBER"
       BEGIN
           CR DUP @ 10 .R SPACE CELL+ DUP @ . CELL+
       DUP CURRENT-ERROR = UNTIL
   THEN
   }}
;

: .REFRESH
    START-MESSAGE
    CURRENT-START 2@ .STATUS
    .STARTS
    .LAST-ACCUMULATED
;

: DISPLAY
    FALSE LOCAL STOPIT

    INIT-CURSOR

    TRUE TO CONNECTED
    GRAPHICS
    PAGE .REFRESH

    BEGIN

       CONNECTED 0= IF
          TRUE TO CONNECTED
          .REFRESH
\          PAGE .REFRESH
       ENDIF

       EKEY CASE

          ESC OF
              FALSE TO CONNECTED
              TRUE TO STOPIT
          ENDOF

          F1 OF
                FALSE TO CONNECTED
                DO-HELP          \ Cycles through help
          ENDOF

          F2 OF
               %>ping PTC CHANNEL-!
          ENDOF

          F3 OF
                FALSE TO CONNECTED
                PAGE .SHOW-ERRORS
                CR ."  Press ESC to return to display"
                BEGIN EKEY ESC = UNTIL
          ENDOF

          --> OF
              FALSE TO CONNECTED
              THE-CURSOR 2@ 1 + THE-CURSOR 2!
          ENDOF

          <-- OF
              FALSE TO CONNECTED
              THE-CURSOR 2@ 1 - THE-CURSOR 2!
          ENDOF

          --^ OF
              FALSE TO CONNECTED
              THE-CURSOR 2@ SWAP 1- SWAP THE-CURSOR 2!
          ENDOF

          --v OF
              FALSE TO CONNECTED
              THE-CURSOR 2@ SWAP 1+ SWAP THE-CURSOR 2!
          ENDOF

          {{ 0 #LL AT-XY ."  Press F1 for help" }}
          BEEP

      ENDCASE
      THE-CURSOR 2@ #LINKS 1- MOD THE-CURSOR 2!
      ADJUST-FIRST-SHOWN
  STOPIT UNTIL
  TEXT
;
