\ COPYRIGHT 1997@1999 DUTCH FORTH WORKSHOP - FILENAME : LEARN.FRT #0

\  Counting twinprimes
\  Copyright 1996 Albert van der Horst SP&AR&C
\  all rights reserved
\
\  learn is about filling in tables.
\ It exposes only `Learn' and some macro's
\ linverse ccor cmult cmask
\ compressedindex lcor lmult lmask

     REVISION -learn " LEARN 2.0 DEBUG and ERROR minimized. "
     NEEDS -twincon
PRIVATES

\ All the mask of i* the current prime
\ Regardless of whether the masks are zero, index on i*p%NPE
NPE HARRAY lmask PRIVATE
\ Given the remainder [i%NPE] it find the index in the tables below
NPE HARRAY linverse PRIVATE
\ Multiples of q (p/210) to skip for next nonzero mask
NPE HARRAY lmult PRIVATE   \ lmask[i-lmult[i]-lcor[i]]!=~0 allways
NPE HARRAY lcor  PRIVATE   \
\ Points to same entry in above table
\ so ccor[compressedindex[i]]=lcor[i] and same for cmult cmask
NPE HARRAY CompressedIndex PRIVATE

\ The next three arrays belong together (FORTRAN trick of parallel arrays)
\ Is not a struct because that doesnot translate to assembly well
\ These are essentially the compressed forms of above tables, i.e.
\ all information bears on masks that really lead to elimination.
\ The nonzero masks of i* the current prime
WP @ CONSTANT cmask PRIMES-PER-ENTRY WALLOT PRIVATE
\ Multiples of q (p/210) to skip for next nonzero mask
WP @ CONSTANT cmult PRIMES-PER-ENTRY WALLOT PRIVATE
\ Correction because q is not precisely p/210
WP @ CONSTANT ccor PRIMES-PER-ENTRY WALLOT  PRIVATE

\ ****************************************************************************/
\                                                                            */
\   Learn how to handle primes that are congruent with ``p'' modulo 210,     */
\   by filling in a lot of tables, that are accessed in the macro's with     */
\   fixed offset.                                                            */
\   In fact p<210 is required, but it is useful to think of p as a prime.    */
\                                                                            */
\ ****************************************************************************/
FWORD pp        PRIVATE         \ Shorthand for p%NPE
FWORD ixn PRIVATE     \ index of numbers in the table, that are divisible by pp  */
FWORD ixnmodNPE PRIVATE
FWORD ixnp PRIVATE    \ index of numbers in the table, previous multiple
                      \ that actually had a bit for it in the table: a hit
FWORD im PRIVATE     \ counts multiples of the prime */
FWORD il PRIVATE      \ index in long table
FWORD ic PRIVATE      \ index in non-compressed table
FWORD incr PRIVATE   \ How far have we to go dwon to find a nonzero mask */

\ Get the mask belonging to case Index a compile time constant
MACRO imm-cgetmask, ( int Index -- )
    cmask + ldl,       \ Fixed location!
END-MACRO

\ Get the mask belonging to case `n' a compile time constant
MACRO cgetmask,   ( FWORD n -- )
    cmask SWAP W[],
END-MACRO

\ Switch from uncompressed tables (index `ixt' ) to compressed
\ ones (`n').
MACRO ToCompressed,
    LOCAL n LOCAL ixt
    CompressedIndex ixt [], n stl,
END-MACRO

\ Leaves a flag that indicates whether the Index (in the mask array)
\ is non trivial, i.e. it would swap a bit.
MACRO non-trivial?,          ( FWORD Index -- )
        lmask ( Index) SWAP [], not,
END-MACRO

\ Leaves a flag that indicates whether the Index (in the mask array)
\ is trivial, i.e. it would not swap a bit.
MACRO trivial?,          ( FWORD Index -- )
        non-trivial?, 0=,
END-MACRO

\ Calculates offset in ENTRY's to arrive at the next nontrivial
\ ENTRY, Index is an index in the arrays of learn
\ that applies to the current ENTRY (for going up in numbers, down in memory)
\ of the target ENTRY (going the other side).
\ bcnt, assumes stride (skipping one ENTRY) is a word.
\ The current ENTRY need not be interesting.
MACRO l-s-nontrivial,       ( FWORD Index -- )
    LOCAL Index
    q ldl,   lmult Index [],   prod,    lcor Index [], add,
END-MACRO

\ With this (byte) stride we skip cmult[n]
\ multiples of p. Starting from an ENTRY index were the compressed
\ index for the mask is n, we find the next higher ENTRY
\ index (lower in memory) by adding what is left in reg A.
MACRO skewstride,        ( FWORD n -- )
        LOCAL n
        q ldl,   cmult n W[],   prod,    ccor n W[], add,
        bcnt,
END-MACRO

\ The ultimate purpose. Used in the fastest loop.
\ As skewstrid, but the compressed index is known and fixed.
MACRO imm-skewstrid,  ( int ICASE -- )
    q ldl,   DUP ( ICASE) cmult + ldl, prod,
    ( ICASE) ccor + ldl, add,
    bcnt,
END-MACRO

\ Correct maskindex `Index' such that it corresponds to the first higher
\ (lower in memory) nontrivial mask.
\ The mask Index points to need not be interesting.
MACRO l-t-nontrivial,       ( FWORD Index -- )
    LOCAL Index
    Index ldl,   lmult Index [], sub, NPE adc, NPE ldc, rem, Index stl,
END-MACRO

\ `Start' is a number index. Split it into a ENTRY-index `ixs' and a
\ mask-index `ixt', i.e. lmask[ixt] can be used to mask the number
\ identified by `Start'
MACRO SplitIndex,
    LOCAL ixt LOCAL ixs LOCAL Start
    0 ldc, Start ldl, NPE ldc,    ldiv, \ Get remainder and quotient
    ( quot) ixs stl,
    ( rem) linverse ldl, wsub, 0 ldnl, ixt stl,
END-MACRO

MACRO LearnMask
    il ldl,                                         \ get
    linverse ixnmodNPE []:=,                         \ store
    MASK-COPY ixnmodNPE [],                               \ get
    lmask il []:=,                                  \ store
END-MACRO PRIVATE

MACRO LearnCorrections
    incr ldl,                                       \ get
    lmult il []:=,                                  \ store
    ixn ldl, NPE ldc, div, ixnp ldl, sub,         \ get
    lcor il []:=,                                  \ store
    il ldl, im stl,
END-MACRO PRIVATE

MACRO LearnIncrements
    incr ldl, 1 adc, incr stl,
    ixn ldl, pp ldl, add, ixn stl,
    ixnmodNPE ldl, pp ldl, add, ixnmodNPE stl,
    ixnmodNPE ldl, NPE 1- ldc, >, IF,
        ixnmodNPE ldl, NPE ldc, sub,   ixnmodNPE stl,
    THEN,
    il ldl, 1 adc, il stl,
END-MACRO PRIVATE

MACRO Learn
    p ldl, NPE ldc, rem, pp stl,
    \  This loop essentially remembers where the next hit is */
    0 ldc, il stl,
    0 ldc, ixn stl,
    0 ldc, ixnmodNPE stl,
    BEGIN,
        LearnMask
        LearnCorrections

        im non-trivial?, IF,
            0 ldc, incr stl,
            ixn ldl, NPE ldc, div, ixnp stl,
        THEN,

        LearnIncrements
    il ldl, NPE 1- ldc, >, UNTIL,

    0 ldc, il stl,
    BEGIN,
        LearnCorrections
        LearnIncrements
    im non-trivial?, UNTIL,

    \  Get nontrivial masks and corresponding attributes into the
    \  other set of tables */
    0 ldc, il stl,      \ Index in normal table
    0 ldc, ic stl,
    BEGIN,
        il non-trivial?, IF,
\             ic ldl,   PRIMES-PER-ENTRY ldc, =, IF,
\                   15 ERROR,
\             THEN,
            lmask il [], cmask ic W[]:=,
            lmult il [], cmult ic W[]:=,
            lcor il [], ccor ic W[]:=,
            \ identify this entry with long table one */
            ic ldl, CompressedIndex il []:=,
            ic ldl, 1 adc, ic stl,
\         ELSE,         \ These indices are never used.
\             NONE ldc, CompressedIndex il []:=,
        THEN,
        il ldl, 1 adc, il stl,
    il ldl, NPE 1- ldc, >, UNTIL,

\     ic ldl,   PRIMES-PER-ENTRY ldc, <>, IF,
\         16 ERROR,
\     THEN,

END-MACRO

DEPRIVE
