\ COPYRIGHT 1997@1999 DUTCH FORTH WORKSHOP - FILENAME : REMOVEP.FRT #0


         REVISION  -removep "--- 2.0B No internal arrays of learn used, only macro's ---"

PRIVATES

\ Points to an entry were we have to mask a bit.

FWORD Pointer    PRIVATE
\  Reset the bits in the sieve for current `p', `m' starting at `Pointer'.
\  Making this a PROC will get this up front, possibly in fast memory.
\ WARNING: uses the fact that the third register remains the same
\ if register a is popped, which is a hidden and gratis dup, .
\ Uses undocumented feature.
PROC SievePrimitive
    p ldl, bcnt, Temp stl,          \ Precalculate stride in address units

    Pointer ldl,
    0 ldc,          \ Balance the 0 from cj, (hidden in UNTIL,)
    \ During the loop there is no descheduling and the pointer
    \ is always in the B-register at the start of the loop.
    \ with 0 in the A-register

\ We thought :
\ Loops run faster, if they start at a word boundary
\       align,                  \ DO NOT DO THIS
\ However this alignment deteroriates 100M from 40.998 to 44.213
\ This is a substantiantal part of the gain of optimisation from 52.476
    BEGIN,
        diff,       \ Get rid of the zero from cj,
        dup,   dup, 0 ldnl,   m ldl,   and,   rev, 0 stnl,
        Temp ldl, add,
    dup,   the-sieve ldl, >, UNTIL,

END-PROC

\  Reset the bits in the sieve for current `p' and `Start'.
\ (A number index in the sieve)
\ But only for the numbers with the same mask as `Start'.
MACRO reset-bits-one-mask   ( Observing Start )

    0 ldc, Start ldl, NPE ldc,    ldiv, \ Get remainder and quotient
    ( quot) neg, the-sieve ldl, wsub,   Pointer stl,
    ( rem) MASK-COPY ldl, wsub,     \ Address of mask needed for the-sieve[start]
    0 ldnl, m stl,              \ Remember mask
    m ldl, not, IF,                        \ Skip if the mask is empty
        SievePrimitive
    ENDIF,
END-MACRO PRIVATE

\   Remove the multiples of `p' from the part of the table that contains
\   `NumberLen' numbers (not ENTRY's). (This number is changed temporarily
\   for this purpose).
\   This is intended and optimal only for `p' being a very small prime and
\   should not be called many times. It does not require that this prime
\   has been `Learn' ed.
\ tested
FWORD Count PRIVATE
MACRO remove-wheel-prime      ( no params )
\       CheckCrc
\           p ldl, DEBUG,

    \ Define `Start ' such that it represents a multiple of `p'`
    last-p-mult

    \ Normally the loop runs NPE times, exceptions for large `p'
    Start  ldl, p ldl, div, 1 adc, Count stl,
    Count ldl,   NPE ldc, >, IF, NPE ldc, Count stl, ENDIF,
    0 ldc, Start ldl, >, IF, #31 ERROR, THEN,

    BEGIN,
        reset-bits-one-mask     \ Uses value of start
        Start  ldl,   p ldl,  sub,  Start  stl,
        Count ldl, -1 adc, Count stl,
        0 ldc, Start ldl, >, IF, 0 ldc, Count stl, THEN,
    1 ldc, Count ldl, >, UNTIL, \ Until zero, but >, is the cheapest comparison
END-MACRO

\ Variables used in both of two following macro's
FWORD ixs  PRIVATE       \ Corresponding entry index into the sieve
FWORD ixt  PRIVATE       \ Corresponding index into the long tables
FWORD n   PRIVATE       \ Corresponding index into the compressed tables
\ The FWORD q is defined in main up front because it is used very often

\ Preparations : calculate `ixs' `ixt' from `Start' and adjust them
\ to point at nontrivial data
MACRO Prepare,
    Start ixs ixt SplitIndex,

    \ In the compressed tables (cmult ccor cmask) we have PRIMES-PER-ENTRY
    \ nontrivial items. Correct the uncompressed situation such that it
    \ is nontrivial.
    ixt trivial?, IF,
        \ Get correction for `ixs' and apply it
        ixt l-s-nontrivial,   neg,   ixs ldl, add, ixs stl,
        \ Correct also `ixt'
        ixt l-t-nontrivial,
     THEN,

END-MACRO  PRIVATE
\   Remove the multiples of `p' from the part of the table that contains
\   `NumberLen' numbers (not ENTRY's). It makes sense that `p' should be a
\   prime number. This is intended and optimal only for `p' being a small
\   prime, say `p'<`len-sieve'. However because this and the following
\   MACRO work for all primes, the choice can be made based on speed. This
\   may depend however from the amount of memory used in sieving, favoring
\   this MACRO in the case of large memory.
\   WARNING: It does require that this prime has been `Learn' ed.
FWORD Counter PRIVATE     \ Over nontrivial masks.
MACRO remove-small-prime      ( no params )
    \ Define `Start' such that it represents a multiple of `p'`
    last-p-mult
\     p ldl, DEBUG,   \ Start ldl, DEBUG,   CheckCrc

    p ldl, NPE ldc, div, q stl,

    Prepare,  ( ics and ixt from Start )

    ixs ldl, -1 ldc, >, IF,
        ixt n ToCompressed, \ Go use `n' , index into compressed tables.

        ixs ldl, neg, the-sieve ldl, wsub,  Pointer stl, \ idem
        PRIMES-PER-ENTRY ldc, Counter stl,
        BEGIN,
            n cgetmask, m stl,   \ Fill in in behalf of SievePrimitive
\           m ldl, DEBUG, n ldl, DEBUG, Pointer ldl, DEBUG,
            SievePrimitive     \ Uses value of `m' and `Pointer'
            \ Next value of `Pointer' using the compressed tables
            n skewstride, Pointer ldl, add, Pointer stl,
            n ldl, -1 adc, n stl,   \ Increment `n' , keeping in proper range
            0 ldc, n ldl, >, IF,
                n ldl, PRIMES-PER-ENTRY adc, n stl,
            THEN,
        Counter ldl, -1 adc, dup, Counter stl, 0=, UNTIL,
    THEN,
\     p ldl, DEBUG,  TDEBUG,

END-MACRO


\ Remove all multiples of the big prime `p' from `the-sieve', It is very
\ efficient where `p' is larger than LEN_SIEVE, or p*210<n_batch, such that
\ no two multiples can be removed using the same mask. This is done by a
\ linear program after some preparation. There are is a pattern of 30
\ removals of bits, because there are 210 or less multiples of p in the
\ table, and none of these hits will have the same residu modulo 210,
\ (because gcd(210,p)=1) so each of the 30 interesting bits in ENTRY's will
\ be hit once, before the pattern repeats. As said above the decision
\ between using `remove-small-prime' and `remove-large-prime' can be based
\ on speed: this will work for all primes.
\ Uses variables in common with `remove-small-prime'
\ They are never used at the same time.
\ This must be a PROC because as soon a all multiples are done a return, is
\ performed. Also this guarantees that this is placed in fast ram, because
\ it is actually assembled immediately.
\ WARNING: It does require that this prime has been `Learn' ed.

\ Address in the jump table are initialised at compile time, hence its
\ allocation in initialised workspace.
WP @ CONSTANT JumpTable PRIMES-PER-ENTRY WALLOT PRIVATE
\ Initialise next JumpTable entry with the current code pointer
: RememberJump ( int ICASE -- )
    CR ." Remembering "
    JumpTable +   $   .S   SWAP WP!
; PRIVATE

\ Assemble the case with index `ICASE' in the compressed tables.
\ Because the location in the table is known, acces to them is optimal.
\ (loop unrolling).
\ Temp contains the entry index a la `ixs'.
MACRO OneCase,   ( int ICASE -- )
    Temp ldl, 0 ldnl,
    DUP ( ICASE) imm-cgetmask, \ Fixed location!
    and,
    Temp ldl, 0 stnl,

    \ Next value of `Temp' using the precalculated tables.
    ( ICASE) imm-skewstrid,   Temp ldl, add, Temp stl,
    Temp ldl, the-sieve ldl, >, IF, return, THEN,
END-MACRO PRIVATE

\ Auxiliary definition: in a PROC we cannot loop, because we
\ are in interpret state.
: 30RegisteredCases
    PRIMES-PER-ENTRY
    BEGIN
       1 -
       DUP RememberJump
       DUP OneCase,
    DUP 0= UNTIL DROP
; PRIVATE

\ Skating down through the sieve using all the prepared variables.
\ This must be a PROC , because in the cases a return is made.
PROC Skating

    ixt n ToCompressed, \ Go use `n' , index into compressed tables.
\     n ldl,  DEBUG,

    ixs ldl, neg, the-sieve ldl, wsub,    Temp stl,
    \ Jump in the middle BEGIN, AGAIN, loop to case n
    JumpTable n W[], gcall,

    \ We need 30 different cases for each of the nontrivial
    \ multiples of p, after that it repeats.
    \ In all of these we use in fact variables, not array indexing
    \ Fallthrough's are essential, this is loop unrolling also

    BEGIN,
        30RegisteredCases               \ Terminates by a return,
    AGAIN,

END-PROC

\ See description above the auxiliary routines.
MACRO remove-large-prime
    \ Define `Start' such that it represents a multiple of `p'`
    last-p-mult
\     p ldl, DEBUG,   \ Start ldl, DEBUG,   CheckCrc

    \ Primes larger than the length of the table?
    \ Sure they are present! And it is important not to waste time on them.
    Start ldl, -1 ldc, >, IF,

        p ldl, NPE ldc, div, q stl,

        Prepare,  ( ics and ixt from Start )

        \ Do not call the proc in the important case that after all
        \ there is no nontrivial multiple to exterminate.
        \ This would be dangerous because Skating tests afterwards.
        ixs ldl, -1 ldc, >, IF,
            Skating
        THEN,

    THEN,
END-MACRO

DEPRIVE
