\ COPYRIGHT 1995@2000 DUTCH FORTH WORKSHOP - FILENAME : TWINCON.FRT #0

 NEEDS -auxtwin

 REVISION -twincon "--constants for twin primes 1.0A using one bit per twin ---"



\ ############## Constants related to the sieves ##################
\ The length of the sieve in ENTRY 's
\ This hasnot much meaning actually.
0 CONSTANT LEN-SIEVE      \ Counting words, dummy

\ The length of the prime table
3500 CONSTANT LEN-BATCH          \ Counting words
LEN-BATCH CELLS CONSTANT BYTES-PER-PRIME-BATCH
BYTES-PER-PRIME-BATCH CONSTANT BPPB    \ Shorthand

DOC
    The first N_SKIP numbers are not investigated.
    The reasons for this is this:
    The first batch must end before N_SKIP*2. Otherwise some primes that
    are used for striking out *must* be used for the end of the table but
    must *not* be used for the beginning. This would add an extra
    complication to the algorithm.

    TWINS_SKIPPED must correspond with N_SKIP, i.e. it must be the number of
    twins that are below N_SKIP.
    Get it from litterature such as Paulo Ribenboim,
    The New Book of Prime record numbers, pg.262.
    Think of NUMBERS_PER_ENTRY as 30 or 210.
ENDDOC
100000 CONSTANT N-SKIP
1224 CONSTANT TWINS-SKIPPED
420 CONSTANT NUMBERS-PER-ENTRY
NUMBERS-PER-ENTRY CONSTANT NPE  \ Abbreviation in behalf of formulae

\ ############## Constants related to prime buffers ###############
 -1 CONSTANT WORTHLESS  \ For offset of prime buffer
 30 CONSTANT PRIMES-PER-BYTE
PRIMES-PER-BYTE CONSTANT PPBY

\ Used to indicate an invalid index.
-1 CONSTANT NONE


\ ############## Constants related to the protocol ################
\ The protocol: > from main to sieve, < the other way
\ %>stop is accepted, but nobody is sending it (yet)
2 CONSTANT %>stop       \ >%
\ Initialise the configuration to a sieve of ``sievesize'', and ``chainlen''
\ more transputers in this line. Each transputer remembers chainlen
\ Pass ``chainlen - 1'' down.
3 CONSTANT %>init       \ >% >W >W        \ sievesize   chainlen
\ Start sieving at offset ``offset''. Pass ``len - sievesize'' down.
4 CONSTANT %>start      \ >% >D >D >$     \ offset, len, masks
\ The last transputer initiates this message, everybody is adding to it.
5 CONSTANT %<twin       \ <% <W           \ result of counting
\ The batch is sent by the parent transputer,
6 CONSTANT %<requestbatch   \ <%  <W         \ request, offset of prime batch
7 CONSTANT %<releasebatch \ <% <W       \ offset of prime batch that was completed
                                        \ or fully sent
8 CONSTANT %>primebatch   \ <% <W <P     \ offset of prime batch, prime batch
\ 9 CONSTANT %<start      \ <% ?? >D ??          \ Left going message used?
10 CONSTANT %>receivebatch     \ <% <W   \ Receive prime batch with
                                        \  indicated offset
11 CONSTANT %>sendcalcbatch     \ <% <W   \ Send prime batch to calculation
12 CONSTANT %>senddownbatch     \ <% <W   \ Send prime batch down
13 CONSTANT %<ack               \ <% <W   \ this communication is ready
14 CONSTANT %>notify            \ %>      \ The second communication task has work.

20 CONSTANT %<error      \ <% <W           \ any feasible debug output
21 CONSTANT %<debug      \ <% <W           \ any feasible debug output
22 CONSTANT %<showstart  \ <% <D           \ (debug output) start of sieving
23 CONSTANT %>ping       \ >%              \ Request to send state information

DOC
\ About the array MASK below:
\ ~MASK[i] contains exactly one bit up for the offsets below
\ Numbers congruent to MAGIC-OFFSET+n(mod NPE) are represented by
the same indicated bit in different entries.
\  All non-primes have an inactive mask, and both of a possible
\ twin pair have the same mask.
\ MAGIC-OFFSET is 130, but this is rather arbitrary.
\ WARNING: a MAGIC-OFFSET of 0 doesnot work because that would
\ make twin primes to span a word in the representation.

\ This table is sent as part of the start message, instead of initialised
\ data. This is because these mask are not used often and in this way
\ they will reside in slow memory.
\ Table below has to agreee with this MAGIC-OFFSET.
\ The table is such that a sieve starting at 10000 (formerly N-SKIP)
\ works nicely.
\    10000 NPE MOD CONSTANT MAGIC-OFFSET
ENDDOC

130 CONSTANT MAGIC-OFFSET
CREATE MASK-IN-MAIN
NPE CELLS ALLOT

\ The flag indicates whether I is relative prime to NPE
: RELATIVE-PRIME  ( int I -- int FLAG )
    NPE GCD 1 =
;

\ Could the number I belong to a twin prime, as far as the relation to
\ NPE is concerned?
: TWIN?   ( INT I -- T/F )
        DUP RELATIVE-PRIME IF
            DUP 2 - RELATIVE-PRIME
            SWAP 2 + RELATIVE-PRIME
            OR
        ELSE
            DROP FALSE
        THEN
;

0 VALUE PRIMES-PER-ENTRY    \ Read : relative primes to NPE
0 VALUE BITS-PER-ENTRY      \ One for each *pair*
\ Fill in MASK-IN-MAIN
: FILL-MASK
     1 LOCAL MASK
     NUMBERS-PER-ENTRY 0 DO
        MAGIC-OFFSET I + TWIN? IF
            1 +TO PRIMES-PER-ENTRY
            MASK
            MAGIC-OFFSET I + 2 +   RELATIVE-PRIME 0= IF
                MASK 1 LSHIFT TO MASK
                1 +TO BITS-PER-ENTRY
            THEN
        ELSE
            0
        THEN
        INVERT
        DUP CR I . H.
        MASK-IN-MAIN I CELLS +   !
    LOOP
;
FILL-MASK
\    END of mask

1 BITS-PER-ENTRY LSHIFT 1 -       \ Contains a bit up for all relevant things.
    CONSTANT TWIN-ALL-SET
DECIMAL
