\ COPYRIGHT 1998@2000 DUTCH FORTH WORKSHOP - FILENAME : TWINHELP.FRT #3


REVISION -twinhelp " 1.0F --- twinhelp : more cosmetics ---"

NEEDS -graphics

 4 CONSTANT #TM     \ Topmargin in characters for display with colors
10 CONSTANT #LM     \ Left margin in pixels
 8 CONSTANT #CHW     \ characters width
16 CONSTANT #CH     \ characters height
49 CONSTANT #SP     \ where the color of a small prime is shown

\ For 2,3,5,7, "prime" "twin"
CREATE COLORS blue , red , green , cyan , yellow , white ,
: COLOR[] CELLS COLORS + @ ;

\ Make a rectangle of SX by SY in color C at position X,Y
: FILLED-RECTANGLE      ( int X,Y,SX,SY,C --)
  LOCAL C LOCAL SY LOCAL SX LOCAL Y LOCAL X
  X SX + X DO
      I Y  I Y SY +  C   DRAW-LINE
  LOOP
;


VARIABLE START-DISPLAY  \ Number displayed at first pos of first line
2100 START-DISPLAY !
VARIABLE END-DISPLAY    \ Number no more displayed
START-DISPLAY @ #LL 1+ #TM - 210 * + END-DISPLAY !
VARIABLE ANIMATED-PRIME \ Prime that is eliminated during the animation
239 ANIMATED-PRIME !

\ Assuming 210 numbers per line, and END-DISPLAY being just off screen,
\ and a line occupying the same height as a character, give the (x,y)
\ coordinates of the number N 's left down corner
: POS-OF-NUM    ( int N -- int X,Y)
    END-DISPLAY @ -
    210 /MOD ( X,Y)
    1+ NEGATE #CH * SWAP
    3 * #LM + SWAP
;

\ Put only the colors instead of the @ in the legenda
: SHOW-LEGEND-COLORS  ( -- )
    6 0 DO
        10 I 4 * + #CHW *
        Ymax #CH -
        #CHW
        #CH
        I COLOR[]
        FILLED-RECTANGLE
    LOOP
;

\ Display a top line with the legenda
\ Colors are filled in later at @
: TOP-LINES
   0 0 AT-XY
   ." Legenda : @ 2 @ 3 @ 5 @ 7 @ P @ T "
   0 1 AT-XY
   ." Small number: divisibility. P : could be prime. T : could be twin"
   0 2 AT-XY
   ." The numbers from " START-DISPLAY ?
   ." to " END-DISPLAY ? ." are represented, 210 per line.
   0 3 AT-XY
   ." We show you the elimination of the multiples of "
   ANIMATED-PRIME ?
   SHOW-LEGEND-COLORS
;

\ Find color as speciifed by legenda
: FIND-COLOR  ( int I -- color C)
        DUP 210 GCD
        DUP 2 MOD 0= IF 2DROP 0 ELSE
        DUP 3 MOD 0= IF 2DROP 1 ELSE
        DUP 5 MOD 0= IF 2DROP 2 ELSE
        DUP 7 MOD 0= IF 2DROP 3 ELSE
        DROP    \ It could be a prime
        \ No twin prime if two less and two more are no prime
        DUP 2 + 210 GCD 1 <>   SWAP 208 + 210 GCD 1 <>  AND IF
            4
        ELSE
            5
        THEN THEN THEN THEN THEN
        COLOR[]
;

\ In a display show the multiples of P in color with index C
\ Fill also in into the legenda
: SHOW-MULTIPLES  \ ( int P, C -- )

    210 0 DO
        #LM I 3 * +   0   3   Ymax #TM #CH * -   \ This is the area
        I FIND-COLOR    \ Color belonging to this relative primeness with 210
        FILLED-RECTANGLE
    LOOP
;

\ Separate the rows of squares that represent numbers by a gray bar
: HORIZONTAL-SEPARATORS
    5 0 DO
        Ymax #CH #TM * - 0 DO
            #LM I J +   Xmax I J +  darkgray DRAW-LINE
        #CH +LOOP
    LOOP
    #CH #CH 5 - DO
        Ymax #CH #TM * - 0 DO
            #LM I J +   Xmax I J +  darkgray DRAW-LINE
        #CH +LOOP
    LOOP
;

\ Show a rectangular display of 210 numbers per line,
\ where the smallest prime divisor determines the color of the position
: SHOW-ALL-MULTIPLES  ( -- )
    SHOW-LEGEND-COLORS  ( -- )
    SHOW-MULTIPLES
    HORIZONTAL-SEPARATORS
;

\ As MS , but a NOOP if there is a keyboard event
: MS/NOOP
    EKEY? IF DROP ELSE MS THEN
;

\ As BEEP , but a NOOP if there is a keyboard event
: BEEP/NOOP
    EKEY? IF ELSE BEEP THEN
;

: HELP-MORE
    CR
    CR ."  Press ESC to return to display, Page Up or Page Down for other help"
;

\ Show how a prime is waste fully eliminated on this display
: SHOW-ELIMINATION
    ANIMATED-PRIME @ LOCAL P
    \ Next higher multiple of P
    END-DISPLAY @ P / P * LOCAL PMULT
    0 LOCAL COLOR
    BEGIN
        PMULT START-DISPLAY @ >= WHILE
        PMULT FIND-COLOR TO COLOR
        63 3 AT-XY ." Eliminating " PMULT . BEEP/NOOP
        \ x y            Size
        3 0 DO
        PMULT POS-OF-NUM 3 16  lightred FILLED-RECTANGLE 250 MS/NOOP
        PMULT POS-OF-NUM 3 16  COLOR FILLED-RECTANGLE 250 MS/NOOP
        LOOP
        500 MS/NOOP
        P NEGATE +TO PMULT
   REPEAT
   3000 MS/NOOP
   0 3 AT-XY 80 SPACES
   0 3 AT-XY ." You see that only a few of the squares hit were white"
   3000 MS/NOOP
   HELP-MORE
;

: .PAGE1

   CR
   CR ."   A twin prime is a pair of primes that have a difference of two."
   CR ."   For example the pair 11,13 is a twin prime, as is 10091,10093."
   CR ."  "
   CR ."   The largest known twin prime is "
   CR ."  "
   CR ."                 570918348 * 10^5120 - 1 and"
   CR ."                 570918348 * 10^5120 + 1 "
   CR ."  "
   CR ."   It is quite probable that there are an infinity of twin primes,"
   CR ."   but nobody has ever managed to prove that."
   CR ."   This program is about counting twin primes under a given number."
   CR ."   For this purpose the range is split up in 'chunks' of"
   CR ."    " CHUNKS D. ."   numbers."
   CR ."   The current chunk is mentioned in the first line of the display."
   CR ."   The large scale shows the overall progress, where the red vertical"
   CR ."   line marks the current world record."
   CR ."   The progress within the chunk is displayed in the small scale beneath "
   CR ."   the large scale."
   CR ."   The chunks are divided in batches where each transputer does a batch,"
   CR ."   whose start is shown in the display."
   CR
   CR ."   Much of the information is from the book by Paolo Ribbenboim:  "
   CR ."   The New Book of Prime Number Records   "
   CR ."   ISBN 0-387-94457-5"
   HELP-MORE
;


: .PAGE2
   CR
   CR ." According to Ribbenboim: "
   CR
   CR ."                        X       prime twins under X
   CR ."                    1,000               35  *)"
   CR ."                   10,000              205  *)"
   CR ."                  100,000            1,224  *)"
   CR ."                1,000,000            8,169  *)"
   CR ."               10,000,000           58,980  *)"
   CR ."              100,000,000          440,312  *)"
   CR ."            1,000,000,000        3,424,506  *)"
   CR ."           10,000,000,000       27,412,679  *)"
   CR ."          100,000,000,000      224,376,048  *)"
   CR ."        1,000,000,000,000    1,870,585,220  *)"
   CR ."       10,000,000,000,000   15,834,664,872  *)"
   CR ."      100,000,000,000,000  135,780,321,665  *)"
   CR ."      137,000,000,000,000  182,312,485,795 current (1998) world record"
   CR
   CR ."  *) Confirmed by this program "
   CR
   HELP-MORE
;

: .PAGE3
    TOP-LINES
    SHOW-ALL-MULTIPLES
    SHOW-ELIMINATION
;

\ Show help about function keys, leave "the user wants more help" into MORE
:    HELP-HELP    ( -- bool MORE)
        PAGE
        CR CR
        ." function key 1 : HELP " CR
        ." function key 2 : PING " CR
        ." function key 3 : ERROR LOG " CR
        ." ESC            : No more display " CR CR

        ." Press function key 1 for more help " CR
        ." Any other key to leave help "
        EKEY F1 =

;

3 CONSTANT MAX-PAGES

0 VALUE CURRENT-PAGE
: DO-HELP
    HELP-HELP LOCAL GOON
    0 TO CURRENT-PAGE

    BEGIN
        GOON WHILE

        \ Display until next key
        CURRENT-PAGE CASE

            0 OF        PAGE .PAGE1             ENDOF
            1 OF        PAGE .PAGE2             ENDOF
            2 OF        PAGE .PAGE3             ENDOF
            3 OF     ." PAGE 3"                 ENDOF
            4 OF     ." PAGE 4"                 ENDOF
            5 OF     ." PAGE 5"                 ENDOF

            BEEP

        ENDCASE

        \ Wait for next key
        EKEY CASE
            ESC OF
                FALSE TO GOON
            ENDOF

            PgDn OF
                1 +TO CURRENT-PAGE
            ENDOF

            F1 OF
                1 +TO CURRENT-PAGE
            ENDOF

            PgUp OF
                -1 +TO CURRENT-PAGE
            ENDOF
            BEEP
        ENDCASE
        CURRENT-PAGE MAX-PAGES MOD TO CURRENT-PAGE   \ Bring in range

  REPEAT
  PAGE
;
